// Source file for the program to test solving systems of equations



// Include files 

#include "RNMath/RNMath.h"



// Program variables

static int solver = RN_CSPARSE_SOLVER;
static int nvariables = 3;
static int exponent = 1;
static int print_verbose = 0;
static int print_debug = 0;



static RNSystemOfEquations *
CreateSystemOfEquations(int nvariables)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate system of equations
  RNSystemOfEquations *system = new RNSystemOfEquations(nvariables);
  if (!system) {
    fprintf(stderr, "Unable to allocate system of equations\n");
    return NULL;
  }

  // Set first variable to 0
  RNPolynomial *polynomial = new RNPolynomial();
  polynomial->AddTerm(1, 0, 1);
  polynomial->AddTerm(0);
  system->InsertEquation(polynomial);

  // Set last variable to 10
  polynomial = new RNPolynomial();
  polynomial->AddTerm(1, nvariables-1, 1);
  polynomial->AddTerm(-10);
  system->InsertEquation(polynomial);
    
  // Set adjacent variables to be equal
  for (int i = 0; i < nvariables-1; i++) {
    RNPolynomial *polynomial = new RNPolynomial();
    polynomial->AddTerm( 1, i,   exponent);
    polynomial->AddTerm(-1, i+1, exponent);
    system->InsertEquation(polynomial);
  }

  // Print the system
  if (print_debug) {
    printf("System:\n");
    system->Print();
  }

  // Print statistics
  if (print_verbose) {
    printf("Created equations ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Variables = %d\n", system->NVariables());
    printf("  # Equations = %d\n", system->NEquations());
    printf("  # Partial Derivatives = %d\n", system->NPartialDerivatives());
    fflush(stdout);
  }

  // Return system of equations
  return system;
}



static int
SolveSystemOfEquations(RNSystemOfEquations *system, int solver)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get convenient variables
  int n = system->NVariables();
  int m = system->NEquations();

  // Allocate data
  RNScalar *x = new RNScalar [ n ];
  RNScalar *y = new RNScalar [ m ];

  // Set initial guess
  for (int i = 0; i < n; i++) x[i] = 0;

  // Solve system of equations
  if (!system->Minimize(x, solver)) {
    fprintf(stderr, "Unable to minimize system of equations\n");
    return 0;
  }

  // Compute residuals
  system->EvaluateResiduals(x, y);

  // Compute error statistics
  RNScalar ssd = 0;
  for (int i = 0; i < m; i++) ssd += y[i] * y[ i ];
  RNScalar rmsd = sqrt(ssd / m);

  // Print statistics
  if (print_verbose) {
    printf("Minimized residuals ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Variables = %d\n", system->NVariables());
    printf("  # Equations = %d\n", system->NEquations());
    printf("  # Partial Derivatives = %d\n", system->NPartialDerivatives());
    printf("  RMSD = %g\n", rmsd);
    fflush(stdout);
  }

  // Print answer
  printf("Solution ...\n");
  for (int i = 0; i < n; i++) 
    printf("  %.3f ", x[i]);
  printf("\n");

  // Delete data
  delete [] x;
  delete [] y;

  // Return success
  return 1;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if (!strcmp(*argv, "-v")) print_verbose = 1; 
    else if (!strcmp(*argv, "-debug")) print_debug = 1; 
    else if (!strcmp(*argv, "-n")) { argc--; argv++; nvariables = atoi(*argv); }
    else if (!strcmp(*argv, "-exponent")) { argc--; argv++; exponent = atoi(*argv); }
    else if (!strcmp(*argv, "-splm")) solver = RN_SPLM_SOLVER; 
    else if (!strcmp(*argv, "-minpack")) solver = RN_MINPACK_SOLVER; 
    else if (!strcmp(*argv, "-ceres")) solver = RN_CERES_SOLVER; 
    else if (!strcmp(*argv, "-csparse")) solver = RN_CSPARSE_SOLVER; 
    else { fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); }
    argv++; argc--;
  }

  // Return OK status 
  return 1;
}



////////////////////////////////////////////////////////////////////////
// MAIN
////////////////////////////////////////////////////////////////////////

int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Create system of equations
  RNSystemOfEquations *system = CreateSystemOfEquations(nvariables);
  if (!system) exit(-1);

  // Solve system of equations
  if (!SolveSystemOfEquations(system, solver)) exit(-1);

  // Delete system of equations
  delete system;

  // Return success 
  return 0;
}











